#include <iostream>
#include <stdexcept>
#include <cctype>
#include "gate.h"

namespace ts {
    map<string, int> servers;
    int servers_inc = 2;
    string servername;

    void con_init()
    {
	servers.clear();
	servers_inc = 2;
	servername.clear();
    }

    void pass(string &prefix, vector<string> &cmd)
    {
	if (cmd.size() == 3 && cmd[2] == "TS") {
	    // PASS pass :TS
	    /*
	     * Send 2.10.4, as it turns off all compatibility fixes for older
	     * 2.10 versions.
	     */
	    van_printf("PASS %s 0210040000 IRC| P\n", cmd[1].c_str());
	} else
	    con_break();
    }
    
    void server(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() == 4) {
	    // :server SERVER name hop comment
	    // announcing new server
	    int &si = servers[cmd[1]];
	    if (!si)
		si = servers_inc++;

	    van_printf(":%s SERVER %s %s %i :%s\n", prefix.c_str(), cmd[1].c_str(),
		    cmd[2].c_str(), si, cmd[3].c_str());
	} else if (!prefix.length() && cmd.size() == 4) {
	    // SERVER name hop comment
	    // handshake
	    servers[cmd[1]] = 1;
	    servername = cmd[1];
	    van_printf("SERVER %s 1 :%s\n", cmd[1].c_str(), cmd[3].c_str());
	} else
	    con_break();
    }

    /*
     * Mangle nicks beginning with a digit
     */
    void manglenick(string &nick)
    {
	if (nick.length() && (nick.length() + 2) <= nicklen && isdigit(nick[0])) {
	    nick.insert(0, "^");
	    while (nick.length() < nicklen)
		nick.append("-");
	}
    }
    
    /*
     * Unmangle nicks
     */
    void unmanglenick(string &nick)
    {
	if (nick.length() > 2 && nick[0] == '^' && *(nick.end() - 1) == '-'
		&& isdigit(nick[1])) {
	    nick.erase(0, 1);
	    nick.erase(nick.find_last_not_of("-") + 1);
	}
    }

    /*
     * Let only UMODE_INVISIBLE | UMODE_OPER | UMODE_WALLOP through
     */
    void fixumode(string &mode)
    {
	for (unsigned int i = 0; i < mode.length();) {
	    switch (mode[i]) {
		case '+':
		case '-':
		case 'i':
		case 'o':
		case 'w':
		    i++;
		    break;
		default:
		    mode.erase(i, 1);
	    }
	}
    }
    
    void nick(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() == 3) {
	    // :nick NICK nick ts
	    // nickchange
	    unmanglenick(prefix);
	    unmanglenick(cmd[1]);

	    van_printf(":%s NICK %s\n", prefix.c_str(), cmd[1].c_str());
	} else if (!prefix.length() && cmd.size() == 9) {
	    // NICK nick hop ts modes username host server comment
	    fixumode(cmd[4]);
	    unmanglenick(cmd[1]);

	    /*
	     * To make realname length consistent, see this in vanilla.cc
	     */
	    if (cmd[8].length() > reallen)
		cmd[8].erase(reallen);

	    van_printf("NICK %s %s %s %s %d %s :%s\n", cmd[1].c_str(),
		    cmd[2].c_str(), cmd[5].c_str(), cmd[6].c_str(),
		    servers[cmd[7]], cmd[4].c_str(), cmd[8].c_str());
	} else
	    con_break();
    }
    
    void sjoin(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() >= 5) {
	    // :server SJOIN ts chan modes [limit] [key] :[@+]nicks
	    int args = ((cmd[3].find('l') != string::npos)?1:0)
		+ ((cmd[3].find('k') != string::npos)?1:0);
	    for (int i = 0; i < args; i++)
		cmd[3] += " " + cmd[4 + i];

	    string nicks, nick;
	    while ((nick = token(cmd[4 + args], ' ')).length()) {
		string nmode(nick, 0, nick.find_first_not_of("@+"));
		nick.erase(0, nmode.length());
		unmanglenick(nick);
		nicks += ((nicks.length())?",":"") + nmode + nick;
	    }

	    van_printf(":%s NJOIN %s :%s\n", prefix.c_str(), cmd[2].c_str(),
		    nicks.c_str());

	    if (cmd[3] != "+" && cmd[3] != "0") {
		van_printf(":%s MODE %s %s\n", prefix.c_str(), cmd[2].c_str(),
			cmd[3].c_str());
	    }
	} else
	    con_break();
    }
    
    void asis(string &prefix, vector<string> &cmd)
    {
	if (prefix.length()) {
	    unmanglenick(prefix);
	    van_printf(":%s ", prefix.c_str());
	}

	for (vector<string>::iterator i = cmd.begin(); i != cmd.end(); i++)
	    van_printf("%s%s%s", (i != cmd.begin())?" ":"",
		    (i != cmd.begin() && (i + 1) == cmd.end())?":":"",
		    i->c_str());

	van_printf("\n");
    }

    void asis_target1(string &prefix, vector<string> &cmd)
    {
	// [:server] NNN target ...
	if (cmd.size() >= 2) {
	    unmanglenick(cmd[1]);
	    asis(prefix, cmd);
	} else
	    con_break();
    }
    
    void pong(string &prefix, vector<string> &cmd)
    {
	// :origin PONG origin destination
	if (cmd.size() >= 3)
	    unmanglenick(cmd[2]);

	asis_target1(prefix, cmd);
    }

    void mode(string &prefix, vector<string> &cmd)
    {
	if (cmd.size() >= 2 && cmd[1][0] != '#') {
	    // :nick MODE nick modes...
	    for (vector<string>::iterator i = cmd.begin() + 2; i != cmd.end(); i++)
		fixumode(*i);

	    asis_target1(prefix, cmd);
	} else if (cmd.size() >= 2 && cmd[1][0] == '#') {
	    // :nick MODE #chan modes...
	    char plus = 0;
	    for (vector<string>::iterator i = cmd.begin() + 2, p = i + 1;
		    i != cmd.end(); i = p++) {
		for (string::iterator j = i->begin(); j != i->end(); j++) {
		    switch (*j) {
			case '+':
			    plus = *j; break;
			case '-':
			    plus = *j; break;
			case 'o':
			case 'v':
			    unmanglenick(*p++); break;
			case 'b':
			case 'e':
			case 'I':
			case 'k':
			    p++; break;
			case 'l':
			    if (plus == '+') p++; break;
		    }
		}
	    }
	    
	    asis(prefix, cmd);
	} else
	    con_break();
    }
    
    void kick(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() >= 3) {
	    // :nick KICK #chan nick [:reason]
	    unmanglenick(cmd[2]);
	    asis(prefix, cmd);
	} else
	    con_break();
    }
    
    void whois(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && (cmd.size() == 2 || cmd.size() == 3)) {
	    // :nick WHOIS [target] nick,nick
	    if (cmd.size() == 3)
		unmanglenick(cmd[1]);
	    string &nicks = *(cmd.end() - 1), tnicks = nicks, nick;
	    nicks.clear();
	    while ((nick = token(tnicks, ',')).length()) {
		unmanglenick(nick);
		nicks += ((nicks.length())?",":"") + nick;
	    }

	    asis(prefix, cmd);
	} else
	    con_break();
    }
    
    void whowas(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() >= 2) {
	    // :nick WHOWAS nick,nick [count [target]]
	    if (cmd.size() >= 3)
		unmanglenick(*(cmd.end() - 1));
	    string &nicks = cmd[1], tnicks = nicks, nick;
	    nicks.clear();
	    while ((nick = token(tnicks, ',')).length()) {
		unmanglenick(nick);
		nicks += ((nicks.length())?",":"") + nick;
	    }

	    asis(prefix, cmd);
	} else
	    con_break();
    }

    void join(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() >= 2) {
	    // :nick JOIN 0 [:nick]
	    if (cmd[1] == "0")
		return asis(prefix, cmd);
	}

	con_break();
    }
    
    void squit(string &prefix, vector<string> &cmd)
    {
	if (cmd.size() == 2) {
	    // :nick SQUIT server
	    cmd.push_back("no reason");
	} else if (cmd.size() == 3 && !cmd[2].length()) {
	    // :nick SQUIT server :
	    cmd[2] = "no reason";
	}

	return asis(prefix, cmd);
    }

    void line(string &prefix, vector<string> &cmd)
    {
	if (cmd[0].length() == 3 && isdigit(cmd[0][0]) && isdigit(cmd[0][1])
		&& isdigit(cmd[0][2]))
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "PASS")
	    pass(prefix, cmd);
	else if (cmd[0] == "SERVER")
	    server(prefix, cmd);
	else if (cmd[0] == "NICK")
	    nick(prefix, cmd);
	else if (cmd[0] == "PING")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "PONG")
	    pong(prefix, cmd);
	else if (cmd[0] == "SJOIN")
	    sjoin(prefix, cmd);
	else if (cmd[0] == "JOIN")
	    join(prefix, cmd);
	else if (cmd[0] == "SQUIT")
	    squit(prefix, cmd);
	else if (cmd[0] == "QUIT")
	    asis(prefix, cmd);
	else if (cmd[0] == "PART")
	    asis(prefix, cmd);
	else if (cmd[0] == "PRIVMSG")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "NOTICE")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "KILL")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "MODE")
	    mode(prefix, cmd);
	else if (cmd[0] == "AWAY")
	    asis(prefix, cmd);
	else if (cmd[0] == "KICK")
	    kick(prefix, cmd);
	else if (cmd[0] == "TOPIC")
	    asis(prefix, cmd);
	else if (cmd[0] == "INVITE")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "MOTD")
	    asis(prefix, cmd);
	else if (cmd[0] == "LUSERS")
	    asis(prefix, cmd);
	else if (cmd[0] == "VERSION")
	    asis(prefix, cmd);
	else if (cmd[0] == "STATS")
	    asis(prefix, cmd);
	else if (cmd[0] == "LINKS")
	    asis(prefix, cmd);
	else if (cmd[0] == "TIME")
	    asis(prefix, cmd);
	else if (cmd[0] == "CONNECT")
	    asis(prefix, cmd);
	else if (cmd[0] == "TRACE")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "ADMIN")
	    asis(prefix, cmd);
	else if (cmd[0] == "INFO")
	    asis(prefix, cmd);
	else if (cmd[0] == "WHOIS")
	    whois(prefix, cmd);
	else if (cmd[0] == "WHOWAS")
	    whowas(prefix, cmd);
	else if (cmd[0] == "USERS")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "WALLOPS")
	    asis(prefix, cmd);
	else if (cmd[0] == "OPERWALL" && (cmd[0] = "WALLOPS").length())
	    asis(prefix, cmd);
	else if (cmd[0] == "ERROR")
	    asis(prefix, cmd);
    }
}
