#include <iostream>
#include <stdexcept>
#include <cctype>
#include "gate.h"

namespace vanilla {
    map<int, string> servers;
    string servername;

    void con_init()
    {
	servers.clear();
	servername.clear();
    }

    void pass(string &prefix, vector<string> &cmd)
    {
	if (cmd.size() >= 4 && string(cmd[2], 0, 3) == "021") {
	    // PASS pass version flags [options]
	    ts_printf("PASS %s :TS\n", cmd[1].c_str());
	    ts_printf("CAPAB :TS EX IE HUB\n");
	} else
	    con_break();
    }
    
    void server(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() == 5) {
	    // :server SERVER name hop token comment
	    // announcing new server
	    servers[atol(cmd[3].c_str())] = cmd[1];

	    ts_printf(":%s SERVER %s %s :%s\n", prefix.c_str(),
		    cmd[1].c_str(), cmd[2].c_str(), cmd[4].c_str());
	} else if (!prefix.length() && cmd.size() == 4) {
	    // SERVER name hop :comment
	    // handshake
	    servers[1] = cmd[1];
	    servername = cmd[1];

	    ts_printf("SERVER %s 1 :%s\n", cmd[1].c_str(), cmd[3].c_str());
	} else if (!prefix.length() && cmd.size() == 5) {
	    // SERVER name hop token :comment
	    // handshake
	    servers[1] = cmd[1];
	    servername = cmd[1];

	    ts_printf("SERVER %s 1 :%s\n", cmd[1].c_str(), cmd[4].c_str());
	} else
	    con_break();
    }
    
    /*
     * Let only UMODE_INVISIBLE | UMODE_OPER | UMODE_WALLOP through
     * Return away status
     */
    int fixumode(string &mode)
    {
	bool plus = true;
	int away = 0;

	for (unsigned int i = 0; i < mode.length();) {
	    switch (mode[i]) {
		case '+':
		    plus = true;
		    i++;
		    break;
		case '-':
		    plus = false;
		case 'i':
		case 'o':
		case 'w':
		    i++;
		    break;
		case 'a':
		    away = 1;
		default:
		    mode.erase(i, 1);
	    }
	}

	return (plus)?away:-away;
    }
    
    void nick(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() == 2) {
	    // :nick NICK nick
	    // nickchange
	    ts::manglenick(prefix);
	    ts::manglenick(cmd[1]);

	    ts_printf(":%s NICK %s :%lu\n", prefix.c_str(), cmd[1].c_str(), 1);
	} else if (!prefix.length() && cmd.size() == 8) {
	    // NICK nick hop username host svtok modes comment
	    int away = fixumode(cmd[6]);
	    ts::manglenick(cmd[1]);

	    /*
	     * Hack around TS servers' warnings about longer realname
	     */
	    if (cmd[7].length() > reallen)
		cmd[7].erase(reallen);

	    ts_printf("NICK %s %s %lu %s %s %s %s :%s\n", cmd[1].c_str(),
		    cmd[2].c_str(), 1, cmd[6].c_str(), cmd[3].c_str(),
		    cmd[4].c_str(), servers[atol(cmd[5].c_str())].c_str(),
		    cmd[7].c_str());
	    if (away > 0)
		ts_printf(":%s AWAY :Gone\n", cmd[1].c_str());
	} else
	    con_break();
    }
    
    void njoin(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() == 3) {
	    // :server NJOIN chan :[@+]nicks
	    if (cmd[1][0] == '!' || cmd[1][0] == '+')
		return;
	    string nicks, nick;
	    while ((nick = token(cmd[2], ',')).length()) {
		string nmode(nick, 0, nick.find_first_not_of("@+"));
		nick.erase(0, nmode.length());
		ts::manglenick(nick);
		nicks += ((nicks.length())?" ":"") + nmode + nick;
	    }

	    ts_printf(":%s SJOIN %lu %s 0 :%s\n", prefix.c_str(), 0,
		    cmd[1].c_str(), nicks.c_str());
	} else
	    con_break();
    }
    
    void part(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() >= 2) {
	    // :nick PART #chan,#chan [reason]
	    ts::manglenick(prefix);
	    string chans, chan;
	    while ((chan = token(cmd[1], ',')).length()) {
		if (chan[0] == '!' || chan[0] == '+')
		    continue;
		chans += ((chans.length())?",":"") + chan;
	    }

	    if (chans.length())
		ts_printf(":%s PART %s%s%s\n", prefix.c_str(),
			chans.c_str(), (cmd.size() >= 3)?" :":"",
			(cmd.size() >= 3)?cmd[2].c_str():"");
	} else
	    con_break();
    }
    
    void asis(string &prefix, vector<string> &cmd)
    {
	if (prefix.length()) {
	    ts::manglenick(prefix);
	    ts_printf(":%s ", prefix.c_str());
	}

	for (vector<string>::iterator i = cmd.begin(); i != cmd.end(); i++)
	    ts_printf("%s%s%s", (i != cmd.begin())?" ":"",
		    (i != cmd.begin() && (i + 1) == cmd.end())?":":"",
		    i->c_str());

	ts_printf("\n");
    }

    void asis_target1(string &prefix, vector<string> &cmd)
    {
	// [:server] NNN target ...
	if (cmd.size() >= 2) {
	    ts::manglenick(cmd[1]);
	    asis(prefix, cmd);
	} else
	    con_break();
    }

    void pong(string &prefix, vector<string> &cmd)
    {
	if (cmd.size() >= 3)
	    ts::manglenick(cmd[2]);

	asis_target1(prefix, cmd);
    }

    void join(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() >= 2) {
	    // :nick JOIN 0 :nick
	    if (cmd[1] == "0")
		return asis(prefix, cmd);

	    // :nick JOIN :#chan\x7Oov,#chan2\x7Oov
	    ts::manglenick(prefix);
	    
	    string chan;
	    while ((chan = token(cmd[1], ',')).length()) {
		if (chan[0] == '!' || chan[0] == '+')
		    continue;
		
		string modes;
		unsigned int c_g = chan.find('\x7');
		if (c_g != string::npos) {
		    for (string::iterator i = chan.begin() + c_g + 1;
			    i != chan.end(); i++) {
			if (*i == 'o')
			    modes += "@";
			else if (*i == 'v')
			    modes += "+";
		    }
		    chan.erase(c_g);
		}

		ts_printf(":%s SJOIN %lu %s 0 :%s%s\n", servername.c_str(), 0,
			chan.c_str(), modes.c_str(), prefix.c_str());
	    }
	} else
	    con_break();
    }

    void msg(string &prefix, vector<string> &cmd)
    {
	if (cmd.size() >= 2) {
	    string chans = cmd[1];
	    while ((cmd[1] = token(chans, ',')).length()) {
		if (cmd[1][0] == '!' || cmd[1][0] == '+')
		    continue;

		asis_target1(prefix, cmd);
	    }
	} else
	    con_break();
    }
    
    void mode(string &prefix, vector<string> &cmd)
    {
	if (cmd.size() >= 2) {
	    /*
	     * 2.10 allows stacking of all this... I'm really not sure if this
	     * will solve it, but we can hope. -lis-
	     *
	     * Huh... It does not really send it, but we should understand it,
	     * so I'm leaving this here... -lis-
	     */
	    string chans = cmd[1];
	    while ((cmd[1] = token(chans, ',')).length()) {
		if (cmd[1][0] == '!' || cmd[1][0] == '+')
		    continue;

		if (cmd[1][0] != '#') {
		    // :nick MODE nick modes...
		    vector<string> tcmd = cmd; // this is ugly, but maybe needed...
		    
		    int away = 0;
		    for (vector<string>::iterator i = tcmd.begin() + 2; i != tcmd.end(); i++)
			away += fixumode(*i);

		    if (!(tcmd.size() == 3 &&
				(tcmd[2] == "+" || tcmd[2] == "-" || tcmd[2].empty())))
			asis_target1(prefix, tcmd);

		    ts::manglenick(cmd[1]);
		    if (away > 0)
			ts_printf(":%s AWAY :Gone\n", cmd[1].c_str());
		    else if (away < 0)
			ts_printf(":%s AWAY\n", cmd[1].c_str());
		} else {
		    // :nick MODE #chan modes...
		    char plus = 0;
		    for (vector<string>::iterator i = cmd.begin() + 2, p = i + 1;
			    i != cmd.end(); i = p++) {
			for (string::iterator j = i->begin(); j != i->end(); j++) {
			    switch (*j) {
				case '+':
				    plus = *j; break;
				case '-':
				    plus = *j; break;
				case 'o':
				case 'v':
				    ts::manglenick(*p++); break;
				case 'b':
				case 'e':
				case 'I':
				case 'k':
				    p++; break;
				case 'l':
				    if (plus == '+') p++; break;
			    }
			}
		    }

		    asis(prefix, cmd);
		}
	    }
	} else
	    con_break();
    }
    
    void kick(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() >= 3) {
	    // :nick KICK #chan,#chan nick,nick [:reason]
	    ts::manglenick(prefix);

	    /*
	     * For each channel kick all users...
	     * That's not what RFC 2812 says, but that's how it's implemented
	     * in 2.10. -lis-
	     */
	    
	    string chan;
	    while ((chan = token(cmd[1], ',')).length()) {
		if (chan[0] == '!' || chan[0] == '+')
		    continue;

		string nicks = cmd[2], nick;
		while ((nick = token(nicks, ',')).length()) {
		    ts::manglenick(nick);
		    ts_printf(":%s KICK %s %s %s%s\n", prefix.c_str(), chan.c_str(),
			    nick.c_str(), (cmd.size() >= 4)?":":"",
			    (cmd.size() >= 4)?cmd[3].c_str():"");
		}
	    }
	} else
	    con_break();
    }
    
    void whois(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && (cmd.size() == 2 || cmd.size() == 3)) {
	    // :nick WHOIS [target] nick,nick
	    if (cmd.size() == 3)
		ts::manglenick(cmd[1]);
	    string &nicks = *(cmd.end() - 1), tnicks = nicks, nick;
	    nicks.clear();
	    while ((nick = token(tnicks, ',')).length()) {
		ts::manglenick(nick);
		nicks += ((nicks.length())?",":"") + nick;
	    }

	    asis(prefix, cmd);
	} else
	    con_break();
    }
    
    void whowas(string &prefix, vector<string> &cmd)
    {
	if (prefix.length() && cmd.size() >= 2) {
	    // :nick WHOWAS nick,nick [count [target]]
	    if (cmd.size() >= 3)
		ts::manglenick(*(cmd.end() - 1));
	    string &nicks = cmd[1], tnicks = nicks, nick;
	    nicks.clear();
	    while ((nick = token(tnicks, ',')).length()) {
		ts::manglenick(nick);
		nicks += ((nicks.length())?",":"") + nick;
	    }

	    asis(prefix, cmd);
	} else
	    con_break();
    }

    void line(string &prefix, vector<string> &cmd)
    {
	if (cmd[0].length() == 3 && isdigit(cmd[0][0]) && isdigit(cmd[0][1])
		&& isdigit(cmd[0][2]))
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "PASS")
	    pass(prefix, cmd);
	else if (cmd[0] == "SERVER")
	    server(prefix, cmd);
	else if (cmd[0] == "NICK")
	    nick(prefix, cmd);
	else if (cmd[0] == "PING")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "PONG")
	    pong(prefix, cmd);
	else if (cmd[0] == "NJOIN")
	    njoin(prefix, cmd);
	else if (cmd[0] == "SQUIT")
	    asis(prefix, cmd);
	else if (cmd[0] == "QUIT")
	    asis(prefix, cmd);
	else if (cmd[0] == "PART")
	    part(prefix, cmd);
	else if (cmd[0] == "JOIN")
	    join(prefix, cmd);
	else if (cmd[0] == "PRIVMSG")
	    msg(prefix, cmd);
	else if (cmd[0] == "NOTICE")
	    msg(prefix, cmd);
	else if (cmd[0] == "KILL")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "MODE")
	    mode(prefix, cmd);
	else if (cmd[0] == "KICK")
	    kick(prefix, cmd);
	else if (cmd[0] == "TOPIC")
	    asis(prefix, cmd);
	else if (cmd[0] == "INVITE")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "MOTD")
	    asis(prefix, cmd);
	else if (cmd[0] == "LUSERS")
	    asis(prefix, cmd);
	else if (cmd[0] == "VERSION")
	    asis(prefix, cmd);
	else if (cmd[0] == "STATS")
	    asis(prefix, cmd);
	else if (cmd[0] == "LINKS")
	    asis(prefix, cmd);
	else if (cmd[0] == "TIME")
	    asis(prefix, cmd);
	else if (cmd[0] == "CONNECT")
	    asis(prefix, cmd);
	else if (cmd[0] == "TRACE")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "ADMIN")
	    asis(prefix, cmd);
	else if (cmd[0] == "INFO")
	    asis(prefix, cmd);
	else if (cmd[0] == "WHOIS")
	    whois(prefix, cmd);
	else if (cmd[0] == "WHOWAS")
	    whowas(prefix, cmd);
	else if (cmd[0] == "USERS")
	    asis_target1(prefix, cmd);
	else if (cmd[0] == "WALLOPS")
	    asis(prefix, cmd);
	else if (cmd[0] == "ERROR")
	    asis(prefix, cmd);
    }
}
